import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns
from math import cos, pi, sin, tan
from scipy.interpolate import interp1d

current_path = os.getcwd()


# Helper functions
def MORELETTERS(i):
    def base10toA(n, A):
        if n < 0:
            raise ValueError("n must be non-negative")
        N = len(A)
        j = n // N
        if j == 0:
            return A[n]
        else:
            return base10toA(j - 1, A) + A[n % N]

    return [base10toA(x - 1, list(string.ascii_uppercase)) for x in i]


def lag0(x):
    return np.concatenate(([0], x[:-1]))


def als(test, yes, no=np.nan):
    return np.where(test, yes, no)


def vul_positie(n, df):  # vul_positie(len(lingra_matrix), f)
    r = np.zeros(n)
    # Ensure indices are integers
    indices = df.iloc[:, 0].astype(int).values - 1
    values = df.iloc[:, 1].values
    r[indices] = values
    return r


def growth_stages(
        Tmean, h, parttable, control, gs
):  # growth_stages(lingra_matrix['G'], h, parttable, control, gs=3)
    # Extractie uit tabel
    BBCH_i = parttable.loc[gs - 1, "BBCH"]
    BBCH_e = parttable.loc[gs - 1, "BBCH_to"]

    TT = np.cumsum(parttable["Thermal_time"]) - np.sum(
        parttable.iloc[0:2, parttable.columns.get_loc("Thermal_time")]
    )
    TT_dpi = TT[gs - 2]  # Thermale tijd drempel
    TT_dpe = TT[gs - 1]  # Thermale tijd drempel

    # Groei graaddagen
    GGD = np.select(
        [
            (Tmean < control.iloc[0]["C92"]) | (Tmean > control.iloc[0]["C95"]),
            (Tmean >= control.iloc[0]["C92"]) & (Tmean <= control.iloc[0]["C93"]),
            (Tmean > control.iloc[0]["C93"]) & (Tmean <= control.iloc[0]["C94"]),
            (Tmean > control.iloc[0]["C94"]) & (Tmean <= control.iloc[0]["C95"]),
        ],
        [
            0,
            (Tmean - control.iloc[0]["C92"]),
            (control.iloc[0]["C93"] - control.iloc[0]["C92"]),
            (control.iloc[0]["C93"] - control.iloc[0]["C92"])
            * (control.iloc[0]["C94"] - Tmean)
            / (control.iloc[0]["C95"] - control.iloc[0]["C94"]),
        ],
        default=0,
    )

    # Creatie van dataframe df_groei
    df_groei = pd.DataFrame({"GGD": GGD})
    df_groei["snede"] = pd.cut(
        np.arange(1, len(df_groei) + 1), bins=[0] + h + [len(df_groei)], labels=False
    )
    df_groei["sumGGD"] = df_groei.groupby("snede")["GGD"].cumsum()
    df_groei["switch_factor"] = np.where(
        (df_groei["sumGGD"] >= TT_dpi) & (df_groei["sumGGD"] < TT_dpe), 1, 0
    )
    df_groei["sumGGDshift"] = df_groei.groupby("snede")["sumGGD"].shift(1, fill_value=0)
    df_groei["gs_on"] = np.where(
        (df_groei["sumGGD"] >= TT_dpi) & (df_groei["sumGGDshift"] < TT_dpe), 1, 0
    )
    df_groei["sumGGD"] = df_groei["sumGGD"] * df_groei["switch_factor"]
    df_groei["BBCHi"] = BBCH_i + (BBCH_e - BBCH_i) * (df_groei["sumGGD"] - TT_dpi) / (
            TT_dpe - TT_dpi
    )
    df_groei["BBCH"] = np.where(df_groei["switch_factor"] != 1, 0, df_groei["BBCHi"])
    df_groei["GGD"] = df_groei["GGD"] * df_groei["gs_on"]
    df_groei.drop(columns=["snede", "BBCHi", "sumGGDshift"], inplace=True)

    return df_groei


def maturation(Tmean, h, parttable, control):
    gs = 7

    BBCH_i = parttable.loc[gs - 2, "BBCH_to"]
    TT_dp = np.cumsum(parttable["Thermal_time"]) - np.sum(
        parttable.iloc[0:2, parttable.columns.get_loc("Thermal_time")]
    )
    TT_dpi = TT_dp[gs - 2]  # thermale tijd drempel
    # Groei graaddagen
    GGD = np.select(
        [
            (Tmean < control.iloc[0]["C92"]) | (Tmean > control.iloc[0]["C95"]),
            (Tmean >= control.iloc[0]["C92"]) & (Tmean <= control.iloc[0]["C93"]),
            (Tmean > control.iloc[0]["C93"]) & (Tmean <= control.iloc[0]["C94"]),
            (Tmean > control.iloc[0]["C94"]) & (Tmean <= control.iloc[0]["C95"]),
        ],
        [
            0,
            (Tmean - control.iloc[0]["C92"]),
            (control.iloc[0]["C93"] - control.iloc[0]["C92"]),
            (control.iloc[0]["C93"] - control.iloc[0]["C92"])
            * (control.iloc[0]["C94"] - Tmean)
            / (control.iloc[0]["C95"] - control.iloc[0]["C94"]),
        ],
        default=0,
    )

    df_groei = pd.DataFrame({"GGD": GGD})
    df_groei["snede"] = pd.cut(
        np.arange(1, len(df_groei) + 1), bins=[0] + h + [len(df_groei)], labels=False
    )
    df_groei["sumGGD"] = df_groei.groupby("snede")["GGD"].cumsum()
    df_groei["switch_factor"] = np.where((df_groei["sumGGD"] >= TT_dpi), 1, 0)
    df_groei["BBCH"] = np.where(df_groei["switch_factor"], BBCH_i, 0)
    df_groei.drop(columns=["GGD", "sumGGD", "snede"], inplace=True)
    return df_groei


def verdeel_irrigatie(i, n=365, imax=10):
    n = 365
    imax = 10
    if i is not None and not i.empty:
        i = i[i["mm"] > 0]

        i_tabel = vul_positie(n, i)

        # Vector maken met gespreide irrigatie
        def calculate_positions(i):
            imax = 10
            posities_list = []

            for _, row in i.iterrows():
                start_day_of_year = row["Dates"]
                mm = row["mm"]
                a = np.ceil(mm / imax).astype(int)
                datums = np.arange(start_day_of_year, start_day_of_year + a * 2, 2)[:a]
                mms = np.concatenate([np.full(a - 1, imax), [mm - (a - 1) * imax]])

                posities_list.append(pd.DataFrame({"datums": datums, "mms": mms}))

            posities = pd.concat(posities_list, ignore_index=True)
            return posities

        posities = calculate_positions(i)

        df_i = pd.DataFrame(data={"mms": vul_positie(365, posities)})
        df_i["rij"] = range(1, len(df_i) + 1)
        # irrigatie uitsmeren tot alle giften onder imax
        while df_i["mms"].max() > imax:
            df_i["rest"] = np.minimum(imax, np.maximum(0, df_i["mms"] - imax))
            df_i["mms_rec"] = df_i["mms"].clip(upper=imax, lower=0)
            df_i["mms"] = df_i["mms_rec"] + df_i["rest"].shift(2, fill_value=0)

        res = df_i["mms"]

    return res if i is not None and not i.empty else 0


def interpol(x, y, fill1, fill2, values):
    f1 = interp1d(x, y, kind="previous", fill_value=(fill1, fill2), bounds_error=False)
    interpol_values = f1(values)
    return interpol_values


def lingra_state(w, h, f, i, s, control, parttable):
    # Create lingra_matrix DataFrame
    lingra_matrix = pd.DataFrame(
        {
            "C": w["Year"],  # Fill weather data
            "AH": w["DOY"],
            "D": w["MAXTMP"],
            "E": w["MINTMP"],
            "H": w["VP"],
            "I": w["RAIN"],
            "J": w["WIND"],
            "K": w["RAD"],
        }
    )

    # Additional calculations
    lingra_matrix["IG"] = lingra_matrix["AH"].isin(f["Dates"]).astype(int)
    lingra_matrix["IH"] = vul_positie(len(lingra_matrix), f)
    lingra_matrix["ND"] = np.where(
        control["C58"] == 2, lingra_matrix["AH"].isin(i["Dates"]).astype(int), 0
    )
    lingra_matrix["NE"] = vul_positie(len(lingra_matrix), i)
    lingra_matrix["AG"] = pd.to_datetime(
        lingra_matrix["AH"] - 1,
        unit="D",
        origin=lingra_matrix["C"][0].astype(str) + "-01-01",
    )
    lingra_matrix["B"] = lingra_matrix["AG"].dt.month
    lingra_matrix["A"] = lingra_matrix["AG"].dt.day
    #lingra_matrix["AK"] = (
    #        2 * pi * (lingra_matrix["AH"] - 80) / w["DOY"].max()
    #)  # Vanaf hier berekeningen
    #lingra_matrix["AL"] = (
    #        0.03358 * (np.sin(2 * pi * lingra_matrix["AH"]) - sin(2 * pi * 80)) / w["DOY"].max()
    #)
    #lingra_matrix["AM"] = lingra_matrix["AK"] + lingra_matrix["AL"]
    #lingra_matrix["AN"] = np.arcsin(0.3978 * np.sin(lingra_matrix["AM"]))
    #lingra_matrix["AO"] = (
    #        -0.10453 / cos(control.iloc[0]["B8"] * pi / 180) * np.cos(lingra_matrix["AN"])
    #)
    #lingra_matrix["AP"] = np.arccos(
    #    lingra_matrix["AO"]
    #    - tan(control.iloc[0]["B8"] * pi / 180) * np.tan(lingra_matrix["AN"])
    #)
    lingra_matrix["F"] = lingra_matrix["D"] - lingra_matrix["E"]
    lingra_matrix["G"] = (lingra_matrix["D"] + lingra_matrix["E"]) / 2 + control.iloc[
        0
    ]["B9"]
    lingra_matrix["R"] = np.where(
        lingra_matrix["G"] <= 3,
        0,
        np.where(
            (lingra_matrix["G"] <= 33) & (lingra_matrix["G"] > 3),
            lingra_matrix["G"] - 3,
            np.where(lingra_matrix["G"] > 33, 30, np.nan),
        ),
    )
    lingra_matrix["V"] = np.maximum(
        0,
        np.minimum(
            control.iloc[0]["C100"] - control.iloc[0]["C99"],
            lingra_matrix["G"] - control.iloc[0]["C99"],
        )
        * control.iloc[0]["C63"],
    )
    lingra_matrix["Q"] = np.maximum(lingra_matrix["G"] - control.iloc[0]["C107"], 0)
    lingra_matrix["S"] = np.maximum(0, lingra_matrix["R"])
    lingra_matrix["W"] = lingra_matrix["D"] - 0.25 * (
            lingra_matrix["D"] - lingra_matrix["E"]
    )
    lingra_matrix["X"] = (
            6.11 * 0.1 * np.exp(17.4 * lingra_matrix["G"] / (lingra_matrix["G"] + 239))
    )
    lingra_matrix["Y"] = np.minimum(lingra_matrix["H"], lingra_matrix["X"])
    lingra_matrix["Z"] = (
            6.10588
            * 0.1
            * np.exp((17.32491 * lingra_matrix["G"]) / (lingra_matrix["G"] + 238.102))
    )
    lingra_matrix["AA"] = (
                                  0.6018 * np.exp((17.27 * lingra_matrix["D"]) / (lingra_matrix["D"] + 237.3))
                                  + 0.6018 * np.exp((17.27 * lingra_matrix["E"]) / (lingra_matrix["E"] + 237.3))
                          ) / 2
    lingra_matrix["AB"] = 101.3 * np.exp(
        -0.034 * control.iloc[0]["C66"] / (lingra_matrix["G"] + 273)
    )
    lingra_matrix["AC"] = (
            239 * 17 * lingra_matrix["X"] / (lingra_matrix["G"] + 239) ** 2
    )
    lingra_matrix["AD"] = (
            4098
            * (0.6108 * np.exp((17.27 * lingra_matrix["G"]) / (lingra_matrix["G"] + 237.3)))
            / (lingra_matrix["G"] + 237.3) ** 2
    )
    lingra_matrix["AE"] = lingra_matrix["Y"] / lingra_matrix["Z"]
    lingra_matrix["AF"] = np.maximum(0, lingra_matrix["G"] - control.iloc[0]["C92"])
    lingra_matrix["AQ"] = lingra_matrix["K"] / 1000
    lingra_matrix["AR"] = lingra_matrix["AQ"] * 0.5
    lingra_matrix["AS"] = lingra_matrix["AH"].isin(h).astype(int)
    lingra_matrix["AT"] = np.where(
        (lingra_matrix["AH"] >= control.iloc[0]["B172"])
        & (lingra_matrix["AH"] <= control.iloc[0]["B173"])
        & (control.iloc[0]["B20"] == 8),
        1,
        0,
    )
    lingra_matrix["P"] = 3
    lingra_matrix["BA"] = np.where(
        control.iloc[0]["C81"] == 0,
        1,
        np.where(np.arange(len(control)) >= control.iloc[0]["C82"], 1, 0),
    )[0]
    lingra_matrix["DN"] = np.cumsum(lingra_matrix["AR"])
    lingra_matrix["DO"] = control.iloc[0]["C97"]
    lingra_matrix["DV"] = np.where(
        (lingra_matrix["G"] - control.iloc[0]["C107"]) <= 0,
        0,
        0.83 * np.log(np.maximum(lingra_matrix["G"], 3)) - 0.8924,
    )
    lingra_matrix["FN"] = np.where(
        (lingra_matrix["G"] <= 50) & (lingra_matrix["G"] >= 30),
        0.05 + (lingra_matrix["G"] - 30) * ((0.09 - 0.05) / (50 - 30)),
        np.where(
            (lingra_matrix["G"] < 30) & (lingra_matrix["G"] >= 15),
            0.02 + (lingra_matrix["G"] - 15) * ((0.05 - 0.02) / (30 - 15)),
            np.where(
                (lingra_matrix["G"] < 15) & (lingra_matrix["G"] >= 10),
                0.01 + (lingra_matrix["G"] - 10) * ((0.02 - 0.01) / (15 - 10)),
                np.where(
                    (lingra_matrix["G"] < 10) & (lingra_matrix["G"] >= -10),
                    (lingra_matrix["G"] + 10) * ((0.02 - 0) / (10 + 10)),
                    np.where(lingra_matrix["G"] == -10, 0, np.nan),
                ),
            ),
        ),
    )
    lingra_matrix["FO"] = control.iloc[0]["C101"]
    lingra_matrix["GG"] = np.where(
        control["B16"] == 40,
        0,
        np.where(
            np.logical_and(control["B16"] > 40, control["B16"] <= 360),
            (control["B16"] - 40) * ((1 - 0) / (360 - 40)),
            np.where(
                np.logical_and(control["B16"] > 360, control["B16"] <= 720),
                1 + (control["B16"] - 360) * ((1.25 - 1) / (720 - 360)),
                np.where(
                    np.logical_and(control["B16"] > 720, control["B16"] <= 1000),
                    1.25 + (control["B16"] - 720) * ((1.35 - 1.25) / (1000 - 720)),
                    np.where(
                        np.logical_and(control["B16"] > 1000, control["B16"] <= 2000),
                        1.35 + (control["B16"] - 1000) * 0,
                        0,
                    ),
                ),
            ),
        ),
    )[0]
    lingra_matrix["GI"] = np.where(
        lingra_matrix["E"] <= 0,
        0,
        np.where(
            np.logical_and(lingra_matrix["E"] >= 3, lingra_matrix["E"] <= 30),
            1,
            ((lingra_matrix["E"] - 0) * ((1 - 0) / (3 - 0))),
        ),
    )
    lingra_matrix["GK"] = np.where(
        (lingra_matrix["AR"] >= 0) & (lingra_matrix["AR"] <= 10),
        1,
        np.where(
            (lingra_matrix["AR"] >= 40) & (lingra_matrix["AR"] <= 42),
            0.33,
            1 + (lingra_matrix["AR"] - 10) * (0.33 - 1) / (40 - 10),
        ),
    )
    lingra_matrix["HD"] = -np.arcsin(
        np.sin(23.45 * control.iloc[0]["C139"])
        * np.cos(2 * control.iloc[0]["C138"] * (lingra_matrix["AH"] + 10) / 365)
    )
    lingra_matrix["HE"] = np.cumsum(lingra_matrix["V"])
    lingra_matrix["HH"] = np.where(
        (control["C60"] <= 0)
        | np.any(lingra_matrix["AH"] == control.iloc[0]["C59"])
        | (control["C60"] > 0) & np.any(lingra_matrix["AH"] == control.iloc[0]["C60"]),
        control.iloc[0]["C59"],
        np.where(control.iloc[0]["C60"] > 0, 0, None),
    )[0]
    lingra_matrix["HI"] = np.where(
        (control["C60"] <= 0)
        | np.any(lingra_matrix["AH"] == control.iloc[0]["C59"])
        | (control["C60"] > 0) & np.any(lingra_matrix["AH"] == control.iloc[0]["C60"]),
        control.iloc[0]["C60"],
        np.where(control.iloc[0]["C60"] > 0, 0, None),
    )[0]
    lingra_matrix["II"] = np.cumsum(lingra_matrix["IH"])
    lingra_matrix["IJ"] = control.iloc[0]["C121"]
    lingra_matrix["KY"] = lingra_matrix["AG"]
    lingra_matrix["KZ"] = 1370 * (
            1 + 0.033 * np.cos(2 * control.iloc[0]["C138"] * lingra_matrix["AH"] / 365)
    )
    lingra_matrix["LA"] = 1 + 0.033 * np.cos((2 * np.pi / 365) * lingra_matrix["AH"])
    lingra_matrix["LB"] = 0.409 * np.sin((2 * np.pi * lingra_matrix["AH"] / 365) - 1.39)
    lingra_matrix["LM"] = control.iloc[0]["C145"] * lingra_matrix["AB"]
    lingra_matrix["LN"] = 0.54 + 0.35 * np.maximum(
        0, np.minimum(1, lingra_matrix["F"] - 12) / 4
    )
    lingra_matrix["MC"] = lingra_matrix["AA"] - lingra_matrix["Y"]
    lingra_matrix["MF"] = np.where(
        control.iloc[0]["B16"] >= 1000,
        0.92,
        np.where(
            (720 <= control.iloc[0]["B16"]) & (control.iloc[0]["B16"] < 1000),
            0.95 + (control.iloc[0]["B16"] - 720) * ((0.92 - 0.95) / (1000 - 720)),
            np.where(
                (360 <= control.iloc[0]["B16"]) & (control.iloc[0]["B16"] < 720),
                1 + (control.iloc[0]["B16"] - 360) * ((0.95 - 1) / (720 - 360)),
                np.where(
                    (40 <= control.iloc[0]["B16"]) & (control.iloc[0]["B16"] < 360),
                    1.05 + (control.iloc[0]["B16"] - 40) * ((1.05 - 1) / (360 - 40)),
                    0,
                ),
            ),
        ),
    )
    lingra_matrix["NO"] = np.where(
        control.iloc[0]["C133"] != 1, s.iloc[0]["mcsat"] - control.iloc[0]["C131"], 0
    )
    lingra_matrix["OE"] = control.iloc[0]["C73"] * lingra_matrix["I"]
    lingra_matrix["OL"] = s.iloc[0]["mcwp"]
    lingra_matrix["ON"] = s.iloc[0]["mcfc"]
    lingra_matrix["OP"] = np.cumsum(lingra_matrix["I"])
    #lingra_matrix["PB"] = np.where(
    #    control.iloc[0]["C61"] == 1,
    #    np.where(
    #        control.iloc[0]["C65"] == 0,
    #        np.where(
    #            lingra_matrix["AH"].shift(1, fill_value=1) - control.iloc[0]["C60"] < 0,
    #            0,
    #            1,
    #        ),
    #        np.where(
    #            365
    #            + lingra_matrix["AH"].shift(1, fill_value=1)
    #            - control.iloc[0]["C60"]
    #            < 0,
    #           0,
    #            1,
    #        ),
    #    ),
    #    0,
    #)
    #lingra_matrix["PD"] = s.iloc[0]["sdpt"]
    #lingra_matrix["AI"] = lingra_matrix["HD"]
    #lingra_matrix["AJ"] = np.where(
    #    control["C60"] <= 0, lingra_matrix["HH"], lingra_matrix["HI"]
    #)
    #lingra_matrix["GH"] = lingra_matrix["GG"]
    #lingra_matrix["GL"] = lingra_matrix["GK"]
    lingra_matrix["HF"] = np.where(
        np.all(
            [
                control.iloc[0]["C61"] == 1,
                np.all(lingra_matrix["HE"] >= control.iloc[0]["C103"]),
                control.iloc[0]["C114"] == 0,
            ]
        ),
        lingra_matrix["AH"],
        control.iloc[0]["C59"],
    )
    lingra_matrix["HG"] = np.where(
        np.all(
            [
                control.iloc[0]["C61"] == 1,
                np.all(lingra_matrix["HE"] >= control.iloc[0]["C103"]),
                control.iloc[0]["C114"] == 0,
            ]
        ),
        lingra_matrix["HF"],
        control.iloc[0]["C59"],
    )
    lingra_matrix["IK"] = lingra_matrix["IH"] * lingra_matrix["IJ"]
    lingra_matrix["LC"] = np.arccos(
        -np.tan((2 * control.iloc[0]["B8"] / 360)) * np.tan(lingra_matrix["LB"])
    )
    lingra_matrix["LF"] = np.sin(
        control.iloc[0]["C139"] * control.iloc[0]["B8"]
    ) * np.sin(lingra_matrix["HD"])
    lingra_matrix["LG"] = np.cos(
        control.iloc[0]["C139"] * control.iloc[0]["B8"]
    ) * np.cos(lingra_matrix["HD"])
    lingra_matrix["LV"] = (
            10
            * 0.26
            * (lingra_matrix["X"] - lingra_matrix["Y"])
            * (0.5 + lingra_matrix["LN"] * lingra_matrix["J"])
    )
    lingra_matrix["LW"] = (
            10
            * 0.26
            * (lingra_matrix["X"] - lingra_matrix["Y"])
            * (1 + lingra_matrix["LN"] * lingra_matrix["J"])
    )
    lingra_matrix["HL"] = np.where(
        np.all(
            [
                control.iloc[0]["C61"] == 1,
                np.all(lingra_matrix["HE"] >= control.iloc[0]["C103"]),
                control.iloc[0]["C114"] == 0,
            ]
        ),
        lingra_matrix["HF"],
        0,
    )
    lingra_matrix["LD"] = (
            ((24 * 60) / np.pi)
            * 0.082
            * lingra_matrix["LA"]
            * (
                    (lingra_matrix["LC"] * np.sin(control.iloc[0]["B8"] * 2 * np.pi / 360))
                    * np.sin(lingra_matrix["LB"])
                    + np.cos(control.iloc[0]["B8"] * 2 * np.pi / 360)
                    * np.cos(lingra_matrix["LB"])
                    * np.sin(lingra_matrix["LC"])
            )
    )
    lingra_matrix["LH"] = 12 * (
            1
            + 2
            * np.arcsin(
        np.maximum(-1, np.minimum(1, lingra_matrix["LF"] / lingra_matrix["LG"]))
        / control.iloc[0]["C138"]
    )
    )
    lingra_matrix["LI"] = 12 * (
            1
            + 2
            * np.arcsin(
        np.maximum(
            -1,
            np.minimum(
                1,
                (-np.sin(-4 * control.iloc[0]["C139"]) + lingra_matrix["LF"])
                / lingra_matrix["LG"],
            ),
        )
        / control.iloc[0]["C138"]
    )
    )
    lingra_matrix["LK"] = np.maximum(
        -1, np.minimum(1, lingra_matrix["LF"] / lingra_matrix["LG"])
    )
    lingra_matrix["PC"] = 1
    lingra_matrix["T"] = lingra_matrix["S"] * lingra_matrix["PC"]
    lingra_matrix["GD"] = np.where(
        np.logical_and(0 <= lingra_matrix["LI"], lingra_matrix["LI"] <= 18), 1, 0.5
    )
    lingra_matrix["LJ"] = 3600 * (
            lingra_matrix["LH"] * lingra_matrix["LF"]
            + 24
            * lingra_matrix["LG"]
            * np.sqrt(1 - lingra_matrix["LK"] * lingra_matrix["LK"])
            / control.iloc[0]["C138"]
    )
    lingra_matrix["LP"] = (
            control.iloc[0]["C140"]
            * (
                    ((lingra_matrix["D"] + 273.16) ** 4 + (lingra_matrix["E"] + 273.16) ** 4)
                    / 2
            )
            * (0.34 - 0.14 * np.sqrt(lingra_matrix["Y"]))
            * (1.35 * lingra_matrix["AQ"] / (0.75 * lingra_matrix["LD"]) - 0.35)
    )
    lingra_matrix["CU"] = np.cumsum(lingra_matrix["T"])
    lingra_matrix["GE"] = lingra_matrix["GD"]
    lingra_matrix["LE"] = (
            np.maximum(0.0001, lingra_matrix["KZ"] * lingra_matrix["LJ"]) / 1000000
    )
    lingra_matrix["LU"] = (1 - control.iloc[0]["C144"]) * lingra_matrix[
        "AQ"
    ] - lingra_matrix["LP"]
    lingra_matrix["GF"] = np.where(
        np.logical_and(control.iloc[0]["C104"] == 1, control.iloc[0]["C62"] == 0),
        lingra_matrix["GE"],
        1,
    )
    lingra_matrix["LL"] = lingra_matrix["AQ"] / lingra_matrix["LE"]
    lingra_matrix["MD"] = (
                                  0.408 * lingra_matrix["AD"] * lingra_matrix["LU"]
                                  + lingra_matrix["LM"]
                                  * (900 / (lingra_matrix["G"] + 273) * lingra_matrix["J"] * lingra_matrix["MC"])
                          ) / (lingra_matrix["AD"] + lingra_matrix["LM"] * (1 + 0.34 * lingra_matrix["J"]))
    lingra_matrix["U"] = lingra_matrix["S"] * lingra_matrix["PC"] * lingra_matrix["GF"]
    lingra_matrix["LO"] = np.maximum(
        0, np.minimum(1, (lingra_matrix["LL"] - 0.2) / 0.56)
    )
    lingra_matrix["CV"] = np.cumsum(lingra_matrix["U"])
    lingra_matrix["LQ"] = (
            control.iloc[0]["C140"]
            * (lingra_matrix["G"] + 273) ** 4
            * (0.56 - 0.079 * np.sqrt(lingra_matrix["Y"] * 10))
            * (0.1 + 0.9 * lingra_matrix["LO"])
    )
    lingra_matrix["HJ"] = np.where(
        np.logical_and(
            control.iloc[0]["C62"] == 0, lingra_matrix["CV"] >= control.iloc[0]["C102"]
        ),
        lingra_matrix["AH"],
        0,
    )
    lingra_matrix["LR"] = (
            lingra_matrix["AQ"] * (1 - control.iloc[0]["C142"]) - lingra_matrix["LQ"]
    )
    lingra_matrix["LS"] = (
            lingra_matrix["AQ"] * (1 - control.iloc[0]["C143"]) - lingra_matrix["LQ"]
    )
    lingra_matrix["LT"] = (
            lingra_matrix["AQ"] * (1 - control.iloc[0]["C144"]) - lingra_matrix["LQ"]
    )
    lingra_matrix["LX"] = (
                                  lingra_matrix["AC"] * (lingra_matrix["LR"] / control.iloc[0]["C141"])
                                  + lingra_matrix["LM"] * lingra_matrix["LV"]
                          ) / (lingra_matrix["AC"] + lingra_matrix["LM"])
    lingra_matrix["LY"] = (
                                  lingra_matrix["AC"] * (lingra_matrix["LS"] / control.iloc[0]["C141"])
                                  + lingra_matrix["LM"] * lingra_matrix["LV"]
                          ) / (lingra_matrix["AC"] + lingra_matrix["LM"])
    lingra_matrix["ME"] = (
                                  lingra_matrix["AC"] * (lingra_matrix["LT"] / control.iloc[0]["C141"])
                                  + lingra_matrix["LM"] * lingra_matrix["LW"]
                          ) / (lingra_matrix["AC"] + lingra_matrix["LM"])
    lingra_matrix["MG"] = np.maximum(0, lingra_matrix["ME"])
    lingra_matrix["MH"] = lingra_matrix["MG"] * lingra_matrix["MF"]
    lingra_matrix["MI"] = lingra_matrix["MG"] * lingra_matrix["MF"]
    lingra_matrix["LZ"] = np.maximum(0, lingra_matrix["LX"])
    lingra_matrix["MA"] = np.maximum(0, lingra_matrix["LY"])
    lingra_matrix["MK"] = (
            1 / (0.76 + 1.5 * (lingra_matrix["MG"] / 10))
            - (5 - control.iloc[0]["C110"]) * 0.1
    )
    lingra_matrix["MP"] = (
            1 / (0.76 + 1.5 * (lingra_matrix["MI"])) - (5 - control.iloc[0]["C110"]) * 0.1
    )
    lingra_matrix["PF"] = lingra_matrix["ME"]
    lingra_matrix["AU"] = 0
    lingra_matrix["AV"] = 0
    lingra_matrix["AW"] = np.where(
        np.logical_and(control.iloc[0]["B18"] == 2, control.iloc[0]["B171"] == 1),
        lingra_matrix["AV"],
        lingra_matrix["AS"],
    )
    lingra_matrix["BB"] = np.where(
        lingra_matrix["BA"] > lingra_matrix["BA"].shift(1, fill_value=0),
        control.iloc[0]["C83"],
        np.where(
            lingra_matrix["AW"] > 0, lingra_matrix["AW"] * control.iloc[0]["C84"], 0
        ),
    )
    lingra_matrix["BC"] = 0
    lingra_matrix["BD"] = 0
    lingra_matrix["BE"] = 0
    lingra_matrix["BF"] = 0
    lingra_matrix["BG"] = 0
    lingra_matrix["BH"] = 0
    lingra_matrix["BI"] = 0
    lingra_matrix["BJ"] = 0
    lingra_matrix["BK"] = 0
    lingra_matrix["BL"] = 0
    lingra_matrix["BM"] = 0
    lingra_matrix["BN"] = 0
    # Combine with growth stages and maturation data
    lingra_matrix = lingra_matrix.join(
        growth_stages(lingra_matrix["G"], h, parttable, control, gs=3).rename(
            columns={
                "switch_factor": "BS",
                "GGD": "BP",
                "sumGGD": "BQ",
                "gs_on": "BO",
                "BBCH": "BT",
            }
        )
    )
    lingra_matrix = lingra_matrix.join(
        growth_stages(lingra_matrix["G"], h, parttable, control, gs=4).rename(
            columns={
                "switch_factor": "BY",
                "GGD": "BV",
                "sumGGD": "BW",
                "gs_on": "BU",
                "BBCH": "BZ",
            }
        )
    )
    lingra_matrix = lingra_matrix.join(
        growth_stages(lingra_matrix["G"], h, parttable, control, gs=5).rename(
            columns={
                "switch_factor": "CE",
                "GGD": "CB",
                "sumGGD": "CC",
                "gs_on": "CA",
                "BBCH": "CF",
            }
        )
    )
    lingra_matrix = lingra_matrix.join(
        growth_stages(lingra_matrix["G"], h, parttable, control, gs=6).rename(
            columns={
                "switch_factor": "CK",
                "GGD": "CH",
                "sumGGD": "CI",
                "gs_on": "CG",
                "BBCH": "CL",
            }
        )
    )
    lingra_matrix = lingra_matrix.join(
        maturation(lingra_matrix["G"], h, parttable, control).rename(
            columns={"switch_factor": "CM", "BBCH": "CN"}
        )
    )

    # Mutations
    lingra_matrix["CP"] = (
            lingra_matrix["BH"]
            + lingra_matrix["BN"]
            + lingra_matrix["BT"]
            + lingra_matrix["BZ"]
            + lingra_matrix["CF"]
            + lingra_matrix["CL"]
            + lingra_matrix["CN"]
    )
    lingra_matrix["CQ"] = interpol(
        parttable["BBCH"], parttable["Total_DM_to_roots"], 0.5, 0, lingra_matrix["CP"]
    )
    lingra_matrix["CR"] = (1 - lingra_matrix["CQ"]) * interpol(
        parttable["BBCH"],
        parttable["Aboveground_to_leaf"],
        1,
        0.05,
        lingra_matrix["CP"],
    )
    lingra_matrix["CS"] = (1 - lingra_matrix["CQ"]) * interpol(
        parttable["BBCH"], parttable["Aboveground_to_stem"], 0, 0.8, lingra_matrix["CP"]
    )
    lingra_matrix["CT"] = (1 - lingra_matrix["CQ"]) * interpol(
        parttable["BBCH"],
        parttable["Aboveground_to_seed"],
        0,
        0.15,
        lingra_matrix["CP"],
    )
    lingra_matrix["DQ"] = np.where(
        np.logical_and(lingra_matrix["CP"] >= 0, lingra_matrix["CP"] <= 90),
        control["C111"],
        control["C111"],
    )
    lingra_matrix["FE"] = np.where(
        np.logical_and(lingra_matrix["CP"] >= 0, lingra_matrix["CP"] <= 90), 0, 0
    )
    lingra_matrix["FH"] = np.where(
        np.logical_and(lingra_matrix["CP"] >= 0, lingra_matrix["CP"] <= 90), 0, 0
    )
    lingra_matrix["FP"] = np.where(
        lingra_matrix["CP"] < lingra_matrix["FO"], 0, lingra_matrix["FN"]
    )
    lingra_matrix["GB"] = np.where(
        control.iloc[0]["B34"] == 1,
        interpol(
            parttable["BBCH"],
            parttable["Aboveground_to_seed"],
            0,
            0.15,
            lingra_matrix["CP"],
        ),
        control.iloc[0]["B38"],
    )

    lingra_matrix["GC"] = lingra_matrix["GB"]

    lingra_matrix["GO"] = np.where(
        lingra_matrix["CP"] <= 60,
        control.iloc[0]["C105"],
        np.where(
            lingra_matrix["CP"] >= 90,
            control.iloc[0]["C106"],
            control.iloc[0]["C105"]
            + (lingra_matrix["CP"] - 60)
            * ((control.iloc[0]["C106"] - control.iloc[0]["C105"]) / (90 - 60)),
        ),
    )

    lingra_matrix["HO"] = np.where(
        (lingra_matrix["CP"] >= 0) & (lingra_matrix["CP"] <= 90),
        control.iloc[0]["C152"],
        control.iloc[0]["C152"],
    )

    lingra_matrix["HP"] = lingra_matrix["HO"]

    lingra_matrix["DR"] = lingra_matrix["DQ"]

    lingra_matrix["HS"] = control.iloc[0]["C154"] * lingra_matrix["HP"]

    lingra_matrix["HT"] = control.iloc[0]["C155"] * lingra_matrix["HP"]

    lingra_matrix["IA"] = control.iloc[0]["C153"] * lingra_matrix["HO"]

    lingra_matrix["DS"] = control.iloc[0]["B22"] / lingra_matrix["DR"]

    lingra_matrix["IC"] = control.iloc[0]["C153"] * lingra_matrix["HT"]

    lingra_matrix["NJ"] = verdeel_irrigatie(i)

    lingra_matrix["NL"] = (1 - control.iloc[0]["C73"]) * lingra_matrix[
        "I"
    ] + lingra_matrix["NJ"]

    lingra_matrix["OQ"] = np.cumsum(lingra_matrix["NJ"])

    lingra_matrix["OR"] = lingra_matrix["OP"] + lingra_matrix["OQ"]

    lingra_matrix["MQ"] = np.cumsum(lingra_matrix["NL"])
    positions = np.where(lingra_matrix["NL"] > 5)[0]
    positions = np.append(
        positions + 1, len(lingra_matrix["NL"])
    )  # Appending True equivalent (len(NL))
    diffs = np.diff(np.concatenate(([0], positions)))
    sequences = np.concatenate([np.arange(1, i + 1) for i in diffs])
    lingra_matrix["NC"] = sequences[: len(lingra_matrix["NL"])]
    lingra_matrix["FU"] = np.where(
        control.iloc[0]["B34"] == 1,
        interpol(
            parttable["BBCH"],
            parttable["Total_DM_to_roots"],
            0.5,
            0,
            lingra_matrix["CP"],
        ),
        control.iloc[0]["B35"],
    )

    lingra_matrix["FY"] = np.where(
        control.iloc[0]["B34"] == 1,
        interpol(
            parttable["BBCH"],
            parttable["Aboveground_to_leaf"],
            1,
            0.05,
            lingra_matrix["CP"],
        ),
        control.iloc[0]["B36"],
    )

    lingra_matrix["FZ"] = np.where(
        control.iloc[0]["B34"] == 1,
        interpol(
            parttable["BBCH"],
            parttable["Aboveground_to_stem"],
            0,
            0.8,
            lingra_matrix["CP"],
        ),
        control.iloc[0]["B37"],
    )

    return lingra_matrix


def approxfun(x, yleft=np.nan, yright=np.nan):
    def interpolator(t):
        if t - 1 < x.keys()[0]:
            return yleft
        elif t - 1 > x.keys()[-1]:
            return yright
        else:
            f = interp1d(
                np.arange(1, x.shape[0]+1),
                x.values,
                bounds_error=False,
                fill_value=(yleft, yright),
            )  # Voorbeeld, gebruik juiste y-waarden
            return f(t).item()

    return interpolator


def case_when(*args, default=None):
    for cond, value in args:
        if cond:
            return value
    return default


# Definieer de als functie
def als(condition, true_val, false_val):
    return true_val if condition else false_val


def lingra_diff(t, parms, dfs):
    print(t)
    state = parms["state"]
    control = parms["control"]
    variable_names = parms["variable_names"]
    y_dict = parms["init"]

    state_f = [approxfun(state[var], yleft=np.nan, yright=np.nan)(t) for var in state]
    state_f_dict = dict(zip(state.keys(), state_f))

    #
    G = state_f_dict["G"]
    MK = state_f_dict["MK"]
    MG = state_f_dict["MG"]
    MI = state_f_dict["MI"]
    GI = state_f_dict["GI"]
    GG = state_f_dict["GG"]
    GK= state_f_dict["GK"]
    #GH = state_f_dict["GH"]
    #GL = state_f_dict["GL"]
    DS = state_f_dict["DS"]
    AW = state_f_dict["AW"]
    FE = state_f_dict["FE"]
    FH = state_f_dict["FH"]
    DV = state_f_dict["DV"]
    IA = state_f_dict["IA"]
    IC = state_f_dict["IC"]
    HO = state_f_dict["HO"]
    HT = state_f_dict["HT"]
    HS = state_f_dict["HS"]
    AH = state_f_dict["AH"]
    HF = state_f_dict["HF"]
    OL = state_f_dict["OL"]
    ON = state_f_dict["ON"]
    CP = state_f_dict["CP"]
    NO = state_f_dict["NO"]
    AR = state_f_dict["AR"]
    CU = state_f_dict["CU"]
    DR = state_f_dict["DR"]
    MA = state_f_dict["MA"]
    NL = state_f_dict["NL"]
    NC = state_f_dict["NC"]
    GO = state_f_dict["GO"]
    FP = state_f_dict["FP"]
    FU = state_f_dict["FU"]
    FY = state_f_dict["FY"]
    OE = state_f_dict["OE"]
    FZ = state_f_dict["FZ"]
    GB = state_f_dict["GB"]
    IK = state_f_dict["IK"]

    C110 = control["C110"]
    C128 = control["C128"]
    C127 = control["C127"]
    C135 = control["C135"]
    C119 = control["C119"]
    C109 = control["C109"]
    C108 = control["C108"]
    B18 = control["B18"]
    B19 = control["B19"]
    C64 = control["C64"]
    C160 = control["C160"]
    C159 = control["C159"]
    C157 = control["C157"]
    C162 = control["C162"]
    C74 = control["C74"]
    C161 = control["C161"]
    C151 = control["C151"]
    C149 = control["C149"]
    C114 = control["C114"]
    C96 = control["C96"]
    C102 = control["C102"]
    C112 = control["C112"]
    C97 = control["C97"]
    B24 = control["B24"]
    B26 = control["B26"]
    C133 = control["C133"]
    C158 = control["C158"]
    C156 = control["C156"]
    C148 = control["C148"]
    C150 = control["C150"]
    C132 = control["C132"]
    C71 = control["C71"]
    B31 = control["B31"]
    C129 = control["C129"]
    C120 = control["C120"]
    C113 = control["C113"]
    C72 = control["C72"]
    C125 = control["C125"]
    C145 = control["C145"]

    if t == 1:
        L = y_dict["L"]
        O = y_dict["O"]
        AY = y_dict["AY"]
        AZ = y_dict["AZ"]
        CY = y_dict["CY"]
        DE = y_dict["DE"]
        DI = y_dict["DI"]
        DU = y_dict["DU"]
        EB = y_dict["EB"]
        EF = y_dict["EF"]
        EH = y_dict["EH"]
        EK = y_dict["EK"]
        ER = y_dict["ER"]
        ES = y_dict["ES"]
        ET = y_dict["ET"]
        FA = y_dict["FA"]
        FG = y_dict["FG"]
        FJ = y_dict["FJ"]
        FL = y_dict["FL"]
        HK = y_dict["HK"]
        HR = y_dict["HR"]
        HV = y_dict["HV"]
        HX = y_dict["HX"]
        IL = y_dict["IL"]
        IQ = y_dict["IQ"]
        JF = y_dict["JF"]
        JJ = y_dict["JJ"]
        JT = y_dict["JT"]
        KA = y_dict["KA"]
        KI = y_dict["KI"]
        KN = y_dict["KN"]
        KO = y_dict["KO"]
        KW = y_dict["KW"]
        MR = y_dict["MR"]
        MV = y_dict["MV"]
        MW = y_dict["MW"]
        MZ = y_dict["MZ"]
        OD = y_dict["OD"]
        OS = y_dict["OS"]
        OT = y_dict["OT"]
        OU = y_dict["OU"]
        OV = y_dict["OV"]
        OX = y_dict["OX"]
        OY = y_dict["OY"]

    else:
        L = dfs[t - 2][45] + dfs[t - 2][1]
        O = dfs[t - 2][46] + dfs[t - 2][2]
        AY = dfs[t - 2][47] + dfs[t - 2][3]
        AZ = dfs[t - 2][48] + dfs[t - 2][4]
        CY = dfs[t - 2][49] + dfs[t - 2][5]
        DE = dfs[t - 2][50] + dfs[t - 2][6]
        DI = dfs[t - 2][51] + dfs[t - 2][7]
        DU = dfs[t - 2][52] + dfs[t - 2][8]
        EB = dfs[t - 2][53] + dfs[t - 2][9]
        EF = dfs[t - 2][54] + dfs[t - 2][10]
        EH = dfs[t - 2][55] + dfs[t - 2][11]
        EK = dfs[t - 2][56] + dfs[t - 2][12]
        ER = dfs[t - 2][57] + dfs[t - 2][13]
        ES = dfs[t - 2][58] + dfs[t - 2][14]
        ET = dfs[t - 2][59] + dfs[t - 2][15]
        FA = dfs[t - 2][60] + dfs[t - 2][16]
        FG = dfs[t - 2][61] + dfs[t - 2][17]
        FJ = dfs[t - 2][62] + dfs[t - 2][18]
        FL = dfs[t - 2][63] + dfs[t - 2][19]
        HK = dfs[t - 2][64] + dfs[t - 2][20]
        HR = dfs[t - 2][65] + dfs[t - 2][21]
        HV = dfs[t - 2][66] + dfs[t - 2][22]
        HX = dfs[t - 2][67] + dfs[t - 2][23]
        IL = dfs[t - 2][68] + dfs[t - 2][24]
        IQ = dfs[t - 2][69] + dfs[t - 2][25]
        JF = dfs[t - 2][70] + dfs[t - 2][26]
        JJ = dfs[t - 2][71] + dfs[t - 2][27]
        JT = dfs[t - 2][72] + dfs[t - 2][28]
        KA = dfs[t - 2][73] + dfs[t - 2][29]
        KI = dfs[t - 2][74] + dfs[t - 2][30]
        KN = dfs[t - 2][75] + dfs[t - 2][31]
        KO = dfs[t - 2][76] + dfs[t - 2][32]
        KW = dfs[t - 2][77] + dfs[t - 2][33]
        MR = dfs[t - 2][78] + dfs[t - 2][34]
        MV = dfs[t - 2][79] + dfs[t - 2][35]
        MW = dfs[t - 2][80] + dfs[t - 2][36]
        MZ = dfs[t - 2][81] + dfs[t - 2][37]
        OD = dfs[t - 2][82] + dfs[t - 2][38]
        OS = dfs[t - 2][83] + dfs[t - 2][39]
        OT = dfs[t - 2][84] + dfs[t - 2][40]
        OU = dfs[t - 2][85] + dfs[t - 2][41]
        OV = dfs[t - 2][86] + dfs[t - 2][42]
        OX = dfs[t - 2][87] + dfs[t - 2][43]
        OY = dfs[t - 2][88] + dfs[t - 2][44]
    dO = (G - O) / 10
    CZ = case_when(
        (O <= 3, 0),
        (3 < O <= 8, (O - 3) * (1 - 0) / (8 - 3)),
        (8 < O <= 30, 1),
        (30 < O <= 35, 1 + (O - 30) * (0 - 1) / (35 - 30)),
        (35 < O <= 40, 0),
    )
    ML = als(
        (C110 < 3)[1],
        MK + (MG - 0.6) / (C110 * (C110 + 3)),
        np.maximum(0.1, np.minimum(0.95, MK)),
    )
    MM = als(
        (C110 < 3)[1],
        MK + (MI / 10 - 0.6) / (C110 * (C110 + 3)),
        np.maximum(0.1, np.minimum(0.95, MK)),
    )
    MN = ((1 - MM) * (C128 - C127) + C127)[1]
    DA = als((CZ <= 0), 0, O * 0.01)
    GJ = CZ * GI
    GP = GJ * GG * GK
    MS = (np.maximum(C127, np.minimum(C135, MR / np.maximum(0.0001, CY))))[
        1
    ]  # pmax(C127, pmin(C135, MR/pmax(0.0001,CY)))
    MX = (np.maximum(C127, np.minimum(C135, MW / np.maximum(0.0001, C119 - CY))))[1]
    DJ = (EF * C109)[1]
    DP = DI / np.maximum(0.1, EB)
    EI = -EH
    EL = EB + EF + EK
    EM = np.maximum(0, EB - DS)
    EO = als(
        (C108 == 1)[1],
        als(((B18 == 1) & (EB >= B19))[1], EF, als(((B18 == 2) & (AW == 1))[1], EF, 0)),
        0,
    )
    EP = als(
        (C108 == 1)[1],
        als(((B18 == 1) & (EB >= B19))[1], EK, als(((B18 == 2) & (AW == 1))[1], EK, 0)),
        0,
    )
    EU = ER + ES + ET
    EV = ER + EB
    FF = EF * FE
    FI = FH * EH
    FM = FL - ER
    GT = (DE * 10000 * (DV * 0.31)) * 0.00000001
    GZ = EF + EB
    HB = FA / C64[1]
    HZ = KA
    IB = IA * EB
    ID = IC * EF
    IO = np.maximum(0, IL - EF * C160)[1]
    IP = IL / als(EF != 0, EF, 1)
    IR = IQ + IL
    IY = np.maximum(0, IQ - EB * C159)[1]
    IZ = np.maximum(HO * EB - IQ, 0)
    JA = IQ / als(EB != 0, EB, 1)
    JB = np.maximum(HT * EF - IL, 0)
    JC = (np.maximum(C157 * EK - JT, 0) / C162)[1]
    JD = JT / als(EK != 0, EK, 1)
    JE = IQ + IL + JT
    JG = np.maximum(HS * EH - JF, 0)
    JO = IQ + IL + JF + JT
    JW = JF / als(EH != 0, EH, 1)
    JY = JF + HV
    if AH > HF:
        KJ = KI / (AH - HF)
    else:
        KJ = KI / (AH + 365 - HF)
    KP = KO + KN
    NF = (CY * (C128 - C127))[1]
    OM = OL * CY
    OO = ON * CY
    EN = als(
        ((B18[1] == 1) & (EB >= B19[1])) | ((B18[1] == 2) & (AW == 1) & (EB > DS)),
        EB - DS,
        0,
    )

    # Niv 1
    MU = (C135[1] - MS) * C74[1]
    MT = (C128[1] - MS) * C74[1]
    NB = (C135[1] - MX) * (C119[1] - C74[1])
    AX = case_when(
        ((B18[1] == 1) & (EU >= B19[1]), -AY),
        (((B18[1] == 2) & (AW == 1) & (EU > DS)), -AY),
        (C108[1] == 0, 0),
        (C108[1] != 0, 1),
    )
    DK = DI + DJ
    DW = als((EN - 0.1) < 0, DV, -DU)
    EQ = EN + EO + EP
    EW = ER + np.maximum(0, EM)
    FC = EB + FM
    FK = EI - FG
    HU = C161[1] * FI
    HW = C160[1] * FF + IP * EO
    IE = (EB * C159[1] + EF * C160[1]) / als(GZ != 0, GZ, 1)
    IF = (IB + ID) / als(GZ != 0, GZ, 1)
    JP = np.minimum((IY + IO) * C151[1], JF - EH * C161[1])
    JX = JE / als(EL != 0, EL, 1)
    KG = IR / als(GZ != 0, GZ, 1)
    KT = MV / OO
    KX = als(CP - C149[1] < 0, als(MR - OM < 0, 0, 1), 0)
    MO = np.maximum(0, np.minimum(1, (MS - C127[1]) / (MN - C127[1])))
    NA = (C128[1] - MX) * (C119[1] - C74[1])
    NP = als(MS >= NO, np.minimum(AH, 4), 0)  # todo not sure lag0 removed by AH
    NQ = np.maximum(0, np.minimum(1, (C135[1] - MS) / (C135[1] - NO)))
    PE = CY * (ON - MS)

    # Niv 2
    DM = als(C114[1] == 1, (1 - np.exp(-C96[1] * DK)) * AR, 0)
    DX = np.maximum(0.01 * (1 + CU / C102[1]), 0.05 * (DK - C97[1]) / C97[1])
    FQ = np.maximum(0, (C112[1] * ((DK - C97[1]) / C97[1])))
    IU = KG * EN
    JQ = JP + IO + IY

    KH = als(
        B24[1] == 0,
        1,
        np.maximum(0.001, np.minimum(1, als((IF - IE) != 0, (KG - IE) / (IF - IE), 1))),
    )
    KL = als(
        B26[1] == 3,
        4000000000 * np.exp(-8400 / (G + 273)),
        als(B26[1] == 4, 4000000000 * np.exp(-8400 / (G + 273)) * (KT), 0),
    )
    KM = als(
        B26[1] == 3,
        5600000000000 * np.exp(-9800 / (G + 273)),
        als(B26[1] == 4, 5600000000000 * np.exp(-9800 / (G + 273)) * KT, 0),
    )
    NR = als(C133[1] != 1, NQ + ((1 - NP / 4) * (1 - NQ)), 1)  # NQ, NP

    #### intermezzo
    dKN = -KN * (1 - np.exp(-KL))
    dKO = -KO * (1 - np.exp(-KM))

    KQ = -dKN
    KR = -dKO
    KS = KQ + KR

    #### niv 3
    DB = KH * 0.693
    DF = DR * np.exp(-C158[1] * (1 - KH))
    DL = DM / AR  # DM
    GM = np.maximum(0, np.minimum(1, 1 - C156[1] * (1.0001 - KH) ** 2))
    IM = als(((C114[1] == 1) & (KH < 1)), EB * C148[1] * (1 - KH), 0)
    IW = np.where(
        (AW == 1) | ((AW == 0) & (B18[1] == 1) & (EN != 0)), IQ - IU, IQ
    ).item()
    JR = als(CP - C150[1] < 0, 0, JQ / C162[1])

    #### niv 4
    DC = als(
        AY < 8,
        np.maximum(0, 0.335 - 0.067 * DK) * CZ,
        np.maximum(0, np.minimum(DB, 0.867 - 0.183 * DK)) * CZ,
    )
    DT = als(((C114[1] == 1) & (KH < 1)), IM * DF, 0)
    GU = als(EN <= 0, GT * (1 / DF), 0)
    JS = np.minimum(JC, JR)
    MB = MA * (1 - DL)
    MJ = np.maximum(0.0001, C132[1] * MI * DL)  # DL

    #### niv 5
    JU = JS * JP / (als(JQ != 0, JQ, 1))
    KC = JS * IO / (als(JQ != 0, JQ, 1))
    KE = JS * IY / (als(JQ != 0, JQ, 1))
    NM = als(
        NL < 0.5,
        MB * np.maximum(0, np.minimum(1, (np.sqrt(NC) - np.sqrt(NC - 1)) * C71[1])),
        0,
    )
    NS = als(
        B31[1] == 0, MJ, np.maximum(0, np.minimum(MV, MO * NR * MJ))
    )  # MV (diff), MO, NR

    #### niv 6
    NN = als(NL >= 5, MB, np.maximum(0, min(MB, NM + NL, (MS - C129[1]))))
    NT = NS / MJ

    #### niv 7
    CX = np.minimum(C120[1] * (als(NT - 0.01 < 0, 0, 1)), C119[1] - CY)  # NT
    FR = (1 - NT) * C113[1]
    FV = als(
        ((NT <= 0) & (NT >= -1)),
        1.6,
        als(((NT < 1) & (NT >= 1.2)), 1, (1.6 + (NT) * ((1 - 1.6) / (1 - 0)))),
    )
    GR = als(
        ((NT <= GM) & (EN == 0)),
        10 * GP * GO * DM * NT,
        als(((NT > GM) & (EN == 0)), 10 * GP * GO * DM * GM, 0),
    )
    IN = np.minimum(NT, KH)
    KU = als(CP - C149[1] < 0, als(NT - 0.01 < 0, 0, 1), 0)
    NW = NL - NN - NS

    #### niv 8
    FS = max(FP, FQ, FR)
    FW = als(B31[1] == 1, FU * FV, FU)
    GN = als(B26[1] == 1, KU, KX)
    GS = GR + FA
    MY = als(C114[1] == 1, CX * MX, 0)
    NX = np.minimum(C72[1] + MU, NW)  # MU, NW
    NZ = als(C114[1] == 1, CX * (MX - C127[1]), 0)  # MX

    #### niv 9
    DY = als(C114[1] == 1, DI * FS, 0)
    GQ = GN * GO * GP
    GV = als(EN <= 0, GT * (1 / DF) * (1 / FY) * (1 / (1 - FW)), 0)
    GY = als(GS - GU <= 0, GS, GU)
    HA = als(C114[1] == 1, DI * FS, 0)
    HY = als(C114[1] == 1, EB * FS, 0)
    KV = als(C114[1] == 1, np.maximum(0, np.minimum(C125[1] * KW * GN, KW)), np.nan)
    NY = als(MT <= NX, np.minimum(C72[1] + NB, NX - MT), 0)  # MT, NB
    OF = OE + np.maximum(0, NW - NX)

    #### niv 10
    DZ = als(EN <= 0, np.minimum((DI - 0.08), DY + DT), EN * DP)  # EN, DI, DY, DT
    FD = als(EN <= 0, np.minimum((DI - 0.08) / DF, HY + IM), EN)
    GW = als(GS - GV <= 0, 0, GS - GV)
    GX = als(GS - GV <= 0, GS, GV)
    HQ = als((B26[1] == 1 | B26[1] == 2), KV, KS)
    OI = NX - NY + MY  # MY
    OA = NX - NY + NZ  # NX, NY, NZ

    #### niv 11
    EG = als(C114[1] == 1, GX * FW - FI, 0)
    EY = GX * FY * (1 - FW)
    HC = GW - HB
    HM = GX * (1 - FW)
    IS = C159[1] * np.maximum(0, FD - EN)

    #### niv 12
    DG = als(C114[1] == 1, DF * EY, 0)  # EY, DF
    ED = als(C114[1] == 1, HM * FY - FD, 0)
    EE = als(C114[1] == 1, HM * FZ - FF, 0)
    EJ = als(C114[1] == 1, HM * GB, 0)
    EZ = EY - FD
    JH = IS

    #### niv 13
    DH = DG - DZ  # samennemen van DI (LAI en RLAI); DH, DZ
    JI = np.maximum(0, IZ + JB + JG - JH)

    #### niv 14
    JK = als(
        C114[1] == 1,
        als(
            B24[1] == 0,
            np.maximum(0, JI) * GN / C64[1],
            (max(0, np.minimum(JI, JJ * GN / C64[1]), 0)),
        ),
        np.nan,
    )

    #### niv 15
    JL = IK / C64[1] - JK + HQ
    JM = als(C114[1] == 1, (JB / (als(JI != 0, JI, 1))) * JK, 0)
    JN = als(C114[1] == 1, (JG / (als(JI != 0, JI, 1))) * JK, 0)
    JZ = als(C114[1] == 1, (IZ / (als(JI != 0, JI, 1))) * JK, 0)
    OB = als(NA <= NY, np.minimum(C72[1], NY - NA), 0)

    #### niv 16
    OJ = NY - OB - MY
    JV = JN - JU - HU
    KD = JM - KC - HW
    KF = JZ - KE - IU
    PA = abs(MW + NY - OB - MY - OY)
    OC = NY - OB - NZ

    #### differentiaalvergelijkingen
    dMR = OI
    dMV = OA
    dCY = CX
    dMW = OJ
    dDI = DH
    dL = DM

    DD = als(DE < 14000, (DC - DX) * DA, -DX * DA)
    dDE = DE * DD

    dAY = AX
    dAZ = case_when(
        ((B18[1] == 1) & (EB > B19[1]), 1),
        (((B18[1] == 2) & (AW == 1) & (EB > DS)), 1),
        default=0,
    )

    dDU = DW
    dEK = EJ - EP
    dER = EN
    dES = EO
    dET = EP
    dFA = HC
    dFG = FF
    dHV = HU
    dHX = HW
    dFJ = FI
    dFL = FD
    dHK = GX
    dHR = -HQ
    dEB = ED
    dEF = EE - EO
    dIL = KD
    dIQ = KF
    dEH = EG
    dJF = JV
    dJJ = JL
    dJT = JS
    dKA = JK
    dKI = als(C114[1] == 1, KH, 0)
    dKW = -KV
    dMZ = MY

    dOD = OC
    OH = dOD

    dOS = OF
    dOT = OB
    dOU = NS
    dOV = NN
    dOX = OI
    dOY = OJ

    s = [
        t,
        dL,
        dO,
        dAY,
        dAZ,
        dCY,
        dDE,
        dDI,
        dDU,
        dEB,
        dEF,
        dEH,
        dEK,
        dER,
        dES,
        dET,
        dFA,
        dFG,
        dFJ,
        dFL,
        dHK,
        dHR,
        dHV,
        dHX,
        dIL,
        dIQ,
        dJF,
        dJJ,
        dJT,
        dKA,
        dKI,
        dKN,
        dKO,
        dKW,
        dMR,
        dMV,
        dMW,
        dMZ,
        dOD,
        dOS,
        dOT,
        dOU,
        dOV,
        dOX,
        dOY,
        L,
        O,
        AY,
        AZ,
        CY,
        DE,
        DI,
        DU,
        EB,
        EF,
        EH,
        EK,
        ER,
        ES,
        ET,
        FA,
        FG,
        FJ,
        FL,
        HK,
        HR,
        HV,
        HX,
        IL,
        IQ,
        JF,
        JJ,
        JT,
        KA,
        KI,
        KN,
        KO,
        KW,
        MR,
        MV,
        MW,
        MZ,
        OD,
        OS,
        OT,
        OU,
        OV,
        OX,
        OY,
        CZ,
        DA,
        DD,
        GJ,
        GP,
        MS,
        MX,
        DJ,
        DP,
        EI,
        EL,
        EM,
        EO,
        EP,
        EU,
        EV,
        FF,
        FI,
        FM,
        GT,
        GZ,
        HB,
        HZ,
        IB,
        ID,
        IO,
        IP,
        IR,
        IY,
        IZ,
        JA,
        JB,
        JC,
        JD,
        JE,
        JG,
        JO,
        JW,
        JY,
        KJ,
        KP,
        NF,
        OM,
        OO,
        EN,
        MU,
        MT,
        NB,
        AX,
        DK,
        DW,
        EQ,
        EW,
        FC,
        FK,
        HU,
        HW,
        IE,
        IF,
        JP,
        JX,
        KG,
        KT,
        KX,
        ML,
        MM,
        MN,
        MO,
        NA,
        NP,
        NQ,
        OH,
        PE,
        DM,
        DX,
        FQ,
        IU,
        JQ,
        KH,
        KL,
        KM,
        NR,
        KQ,
        KR,
        KS,
        DB,
        DF,
        DL,
        GM,
        IM,
        IW,
        JR,
        DC,
        DT,
        GU,
        JS,
        MB,
        MJ,
        JU,
        KC,
        KE,
        NM,
        NS,
        NN,
        NT,
        CX,
        FR,
        FV,
        GR,
        IN,
        KU,
        NW,
        FS,
        FW,
        GN,
        GS,
        MY,
        NX,
        NZ,
        DY,
        GQ,
        GV,
        GY,
        HA,
        HY,
        KV,
        NY,
        OF,
        DZ,
        FD,
        GW,
        GX,
        HQ,
        OI,
        OA,
        EG,
        EY,
        HC,
        HM,
        IS,
        DG,
        ED,
        EE,
        EJ,
        EZ,
        JH,
        DH,
        JI,
        JK,
        JL,
        JM,
        JN,
        JZ,
        OB,
        OJ,
        JV,
        KD,
        KF,
        PA,
        OC,
        AH,
        HF,
    ]
    # Output
    return s


def lees_control_params():
    # Read control parameters from CSV file
    controls_path = os.path.join(current_path, "grassat/input/lingra/controls.csv")
    df_control = pd.read_csv(controls_path, header=None)

    # Transpose and set column names
    control = df_control.iloc[:, 1:].T
    control.columns = df_control.iloc[:, 0]

    # Set specific control parameter values
    control["B8"] = 52  # latitude
    control["B9"] = 0  # T (°C) change from default
    control["B15"] = 0  # (ppm) Increase in carbon dioxide from default
    control["B16"] = 417  # (ppm) Assumed CO2 concentration
    control["B22"] = 2  # Leaf area index after cutting (CLAI)
    control["B24"] = (
        1  # Nitrogen stress included? (0 = no nitrogen effects; 1 = nitrogen effects included)
    )
    control["B26"] = (
        4
        # (1) Default LINGRA-N (based on transp. factor); (2) Moisture dependent; (3) Temperature dependent; (4) Moisture & temp dependent
    )
    control["B27"] = 75  # Total mineral soil N available at start of growth period
    control["B29"] = 400  # (mm) Root depth
    control["B31"] = (
        1  # Is drought stress included? (0 = no drought effects; 1 = drought effects included)
    )
    control["B34"] = (
        1  # Is partitioning fixed or varied (0 = fixed; 1 = automatic (See B85:G91)
    )
    control["C114"] = 1  # Emergence
    control["C133"] = 0  # Air ducts in roots present (1) or not (0)
    control["C66"] = 23  # Elevation
    control["B16"] = 400  # CO2

    return control


def lees_parttable():
    # Read partitioning table from CSV file
    partitioning_path = os.path.join(current_path, "grassat/input/lingra/partitioning_table.csv")
    return pd.read_csv(partitioning_path)


def initial_params(w, control, parttable):
    # Initialize parameters
    CPi = 21
    CYi = control["C74"][1]
    ONi = control["C128"][1]

    MSi = max(control["C127"][1], min(control["C69"][1], control["C128"][1]))
    MVi = control["C74"][1] * (MSi - control["C127"][1])
    OOi = ONi * CYi

    Gi = (w["MINTMP"][0] + w["MAXTMP"][0]) / 2 + control["B9"][1]
    KTi = MVi / OOi
    KLi = np.where(
        control["B26"][1] == 3,
        4000000000 * np.exp(-8400 / (Gi + 273)),
        np.where(
            control["B26"][1] == 4, 4000000000 * np.exp(-8400 / (Gi + 273)) * KTi, 0
        ),
    )
    KMi = np.where(
        control["B26"][1] == 3,
        5600000000000 * np.exp(-9800 / (Gi + 273)),
        np.where(
            control["B26"][1] == 4, 5600000000000 * np.exp(-9800 / (Gi + 273)) * KTi, 0
        ),
    )

    KRi = control["C123"][1] * (1 - np.exp(-KMi))
    KQi = control["C122"][1] * (1 - np.exp(-KLi))
    KSi = KQi + KRi
    JJi = np.where(
        np.isin(control["B26"][1], [1, 2]), control["C124"][1], control["C124"][1] + KSi
    )

    HOi = np.where((CPi >= 0) & (CPi <= 90), control["C152"][1], control["C152"][1])
    HSi = control["C154"][1] * HOi
    HTi = control["C155"][1] * HOi
    KNi = control["C122"][1]
    KOi = control["C123"][1]
    KPi = KOi + KNi
    ERi = 0
    FLi = 0
    FMi = FLi - ERi
    FZi = np.where(
        control["B34"][1] == 1,
        np.interp(21, parttable.iloc[:, 0], parttable.iloc[:, 3]),
        control["B37"][1],
    )
    FUi = np.where(
        control["B34"][1] == 1,
        np.interp(21, parttable.iloc[:, 0], parttable.iloc[:, 5]),
        control["B35"][1],
    )
    EHi = FUi * control["C98"][1]
    ECi = control["C98"][1] - EHi
    GBi = np.where(
        control["B34"][1] == 1,
        np.interp(CPi, parttable.iloc[:, 0], parttable.iloc[:, 4]),
        control["B38"][1],
    )
    MXi = max(control["C127"][1], min(control["C70"][1], control["C128"][1]))
    FXi = np.where(
        control["B34"][1] == 1,
        np.interp(CPi, parttable.iloc[:, 0], parttable.iloc[:, 2]),
        control["B36"][1],
    )

    # Initial values for differential equations
    Li = 0
    Oi = control["C78"][1]
    AYi = 0
    AZi = 0
    CYi = control["C74"][1]
    DEi = control["C76"][1]
    DIi = control["C75"][1]
    DUi = 0
    EBi = FXi * ECi
    EFi = FZi * ECi
    EHi = FUi * control["C98"][1]
    EKi = GBi * ECi
    ERi = 0
    ESi = 0
    ETi = 0
    FAi = control["C77"][1]
    FGi = 0
    FJi = 0
    FLi = 0
    HKi = ERi + EBi + FMi + ESi + EFi + EKi + EHi
    KWi = control["C124"][1]
    HRi = np.where(
        (control["B26"][1] == 1) | (control["B26"][1] == 2), KWi, KPi
    )  # !!! change to R code KWi variable is put firs, otherwise an error
    HVi = 0
    HXi = 0
    ILi = HTi * EFi
    IQi = HOi * EBi
    JFi = HSi * EHi
    JTi = control["C157"][1] * EKi
    KAi = 0
    KIi = 0
    KNi = control["C122"][1]
    KOi = control["C123"][1]
    MRi = MSi * CYi
    MVi = control["C74"][1] * (MSi - control["C127"][1])
    MWi = MXi * (control["C119"][1] - CYi)
    MZi = 0
    ODi = (control["C119"][1] - control["C74"][1]) * (MXi - control["C127"][1])
    OSi = 0
    OTi = 0
    OUi = 0
    OVi = 0
    OXi = 0
    OYi = 0

    # Define initial values
    init = {
        "L": Li,
        "O": Oi,
        "AY": AYi,
        "AZ": AZi,
        "CY": CYi,
        "DE": DEi,
        "DI": DIi,
        "DU": DUi,
        "EB": EBi,
        "EF": EFi,
        "EH": EHi,
        "EK": EKi,
        "ER": ERi,
        "ES": ESi,
        "ET": ETi,
        "FA": FAi,
        "FG": FGi,
        "FJ": FJi,
        "FL": FLi,
        "HK": HKi,
        "HR": HRi.item(0),
        "HV": HVi,
        "HX": HXi,
        "IL": ILi,
        "IQ": IQi,
        "JF": JFi,
        "JJ": JJi.item(0),
        "JT": JTi,
        "KA": KAi,
        "KI": KIi,
        "KN": KNi,
        "KO": KOi,
        "KW": KWi,
        "MR": MRi,
        "MV": MVi,
        "MW": MWi,
        "MZ": MZi,
        "OD": ODi,
        "OS": OSi,
        "OT": OTi,
        "OU": OUi,
        "OV": OVi,
        "OX": OXi,
        "OY": OYi,
    }

    return init


def lingra(w, h, f, i, s, control=None):
    # Control parameters and partitioning table reading
    X = lees_control_params()
    parttable = lees_parttable()

    # Read or extend control parameters
    if control is None:
        control = X
    else:
        control = np.hstack((X[:, ~np.isin(X.columns, control.columns)], control))

    # Get initial values for differential equations
    init = initial_params(w, control, parttable)

    # State variables iteration
    state = lingra_state(w, h, f, i, s, control, parttable)

    # Solve differential equations
    #t_span = (1, range(1, w["DOY"].max()))
    #t_eval = range(1,  w["DOY"].max()+1)
    y_list = list(init.values())
    variable_names = list(init.keys())
    parms = {
        "state": state,
        "control": control,
        "variable_names": variable_names,
        "init": init,
    }
    # Assuming simulation for 365 days
    dfs = []
    for t in range(1,  w["DOY"].max()+1):
        y = lingra_diff(t, parms, dfs)
        dfs.append(y)
    columns = [
        "t",
        "dL",
        "dO",
        "dAY",
        "dAZ",
        "dCY",
        "dDE",
        "dDI",
        "dDU",
        "dEB",
        "dEF",
        "dEH",
        "dEK",
        "dER",
        "dES",
        "dET",
        "dFA",
        "dFG",
        "dFJ",
        "dFL",
        "dHK",
        "dHR",
        "dHV",
        "dHX",
        "dIL",
        "dIQ",
        "dJF",
        "dJJ",
        "dJT",
        "dKA",
        "dKI",
        "dKN",
        "dKO",
        "dKW",
        "dMR",
        "dMV",
        "dMW",
        "dMZ",
        "dOD",
        "dOS",
        "dOT",
        "dOU",
        "dOV",
        "dOX",
        "dOY",
        "L",
        "O",
        "AY",
        "AZ",
        "CY",
        "DE",
        "DI",
        "DU",
        "EB",
        "EF",
        "EH",
        "EK",
        "ER",
        "ES",
        "ET",
        "FA",
        "FG",
        "FJ",
        "FL",
        "HK",
        "HR",
        "HV",
        "HX",
        "IL",
        "IQ",
        "JF",
        "JJ",
        "JT",
        "KA",
        "KI",
        "KN",
        "KO",
        "KW",
        "MR",
        "MV",
        "MW",
        "MZ",
        "OD",
        "OS",
        "OT",
        "OU",
        "OV",
        "OX",
        "OY",
        "CZ",
        " DA",
        " DD",
        "GJ",
        "GP",
        "MS",
        "MX",
        "DJ",
        "DP",
        "EI",
        "EL",
        "EM",
        "EO",
        "EP",
        "EU",
        "EV",
        "FF",
        "FI",
        "FM",
        "GT",
        "GZ",
        "HB",
        "HZ",
        "IB",
        "ID",
        "IO",
        "IP",
        "IR",
        "IY",
        "IZ",
        "JA",
        "JB",
        "JC",
        "JD",
        "JE",
        "JG",
        "JO",
        "JW",
        " JY",
        "KJ",
        "KP",
        "NF",
        "OM",
        "OO",
        "EN",
        "MU",
        "MT",
        "NB",
        "AX",
        "DK",
        "DW",
        "EQ",
        "EW",
        "FC",
        "FK",
        "HU",
        "HW",
        "IE",
        "IF",
        "JP",
        "JX",
        "KG",
        "KT",
        "KX",
        "ML",
        "MM",
        "MN",
        "MO",
        "NA",
        "NP",
        "NQ",
        "OH",
        "PE",
        "DM",
        "DX",
        " FQ",
        "IU",
        "JQ",
        "KH",
        "KL",
        "KM",
        "NR",
        " KQ",
        "KR",
        "KS",
        "DB",
        "DF",
        "DL",
        "GM",
        "IM",
        "IW",
        "JR",
        "DC",
        "DT",
        "GU",
        "JS",
        "MB",
        "MJ",
        "JU",
        "KC",
        "KE",
        "NM",
        "NS",
        "NN",
        "NT",
        "CX",
        "FR",
        "FV",
        "GR",
        "IN",
        "KU",
        "NW",
        "FS",
        "FW",
        "GN",
        "GS",
        "MY",
        "NX",
        "NZ",
        "DY",
        "GQ",
        "GV",
        "GY",
        "HA",
        "HY",
        "KV",
        "NY",
        "OF",
        "DZ",
        "FD",
        "GW",
        "GX",
        "HQ",
        "OI",
        "OA",
        "EG",
        "EY",
        "HC",
        "HM",
        "IS",
        "DG",
        "ED",
        "EE",
        "EJ",
        "EZ",
        "JH",
        "DH",
        "JI",
        "JK",
        "JL",
        "JM",
        "JN",
        "JZ",
        "OB",
        "OJ",
        "JV",
        "KD",
        "KF",
        "PA",
        "OC",
        "AH",
        "HF",
    ]
    final = pd.DataFrame(dfs, columns=columns)
    # Additional calculations
    final["OG"] = final["MV"][0] - final["MV"]
    final["IV"] = np.cumsum(final["IU"])
    final["KB"] = final["IV"] + final["IW"] + final["HX"] + final["IL"] + final["JF"]
    final["KK"] = (
            final["KA"]
            + (final["IQ"] + final["IL"] + final["JF"] + final["JT"])
            - (
                    final["IQ"]
                    + final["IL"]
                    + final["JF"]
                    + final["JT"]
                    + final["IV"]
                    + final["HV"]
                    + final["HX"]
            )
    )
    final["IX"] = final["HX"] + final["HV"] + final["IV"]
    final["OK"] = final["OG"] + final["OH"]
    final["NU"] = np.cumsum(final["NT"])
    final["NV"] = np.where(
        final["AH"] > final["HF"],
        final["NU"] / (final["AH"] - final["HF"]),
        final["NU"] / (final["AH"] + 365 - final["HF"]),
    )
    final["OW"] = np.cumsum(final["OI"])
    final["IT"] = np.cumsum(final["IS"])
    final["EC"] = final["EL"] + final["FM"] + final["FG"]
    final.loc[0, "EC"] = (control["C98"] - final["EH"][0])[1]
    final["EX"] = final["ER"] + final["EC"]
    final["FB"] = final["ER"] + final["EC"]
    final["M"] = (final["EX"] / 10) / np.where(final["L"] != 0, final["L"], 1)
    FINAL = pd.concat([final, state], axis=1)
    #FINAL = FINAL.loc[FINAL["t"] <= h[-1]]    # output only untill final cut
    return FINAL


def lingra_wig_response(lingra_data):
    kols = {
        'EF': 'Weight of living stems; WST',
        'EM': 'Harvestable leaf weight; HRVBL',
        'AG': "Date",
        'AS': 'Harvest date',
        'IG': 'Fertilizer N application; (0=NO, 1=YES)',
        "KG": "N concentration in total vegetative living above ground biomass",
        "EQ": "Total harvested weight"
    }
    response = lingra_data[kols.keys()].copy()
    response.columns = kols.values()
    response['Date'] = pd.to_datetime(response['Date'], unit='D')
    return response.to_json()


def lingraplot_grass_yield(lingra_data, outdir):
    # Define the column names
    os.makedirs(outdir, exist_ok=True)
    kols = {
        "EF": "Weight of living stems; WST",
        "EM": "Harvestable leaf weight; HRVBL",
        "AG": "Date",
        "AS": "Harvest date",
        "IG": "Fertilizer N application; (0=NO, 1=YES)",
    }

    stems = "Weight of living stems; WST"
    leaves = "Harvestable leaf weight; HRVBL"

    m = max(lingra_data["EF"] + lingra_data["EM"]) * 1.1

    bollen = lingra_data[kols.keys()].copy()
    bollen.columns = kols.values()
    bollen["Date"] = pd.to_datetime(bollen["Date"], unit="D")
    bollen = bollen.rename(
        columns={
            "Harvest date": "maai",
            "Fertilizer N application; (0=NO, 1=YES)": "stikstof",
        }
    )
    bollen = bollen[(bollen["maai"] == 1) | (bollen["stikstof"] == 1)]
    bollen["maai"] = np.where(bollen["maai"] == 1, bollen["maai"] * m, np.nan)
    bollen["stikstof"] = np.where(
        bollen["stikstof"] == 1, bollen["stikstof"] * m * 1.05, np.nan
    )

    p = lingra_data[kols.keys()].copy()
    p.columns = kols.values()
    p["Date"] = pd.to_datetime(p["Date"], unit="D")
    p = p.rename(
        columns={
            "Harvestable leaf weight; HRVBL": "Blad",
            "Weight of living stems; WST": "Stengel",
        }
    )
    p["stengel en blad"]=p["Blad"]+p["Stengel"]
    p["Harvest date"] = p["Harvest date"].shift(1, fill_value=0)
    p["periode"] = p["Harvest date"].cumsum()
    p.to_csv(os.path.join(outdir, 'gras_plot_biomass.csv'), index=False)
    p_Rcomp = p.melt(
        id_vars=["Date", "periode"],
        value_vars=["Blad", "Stengel"],
        var_name="AGB_var",
        value_name="AGB_value",
    )

    plt.figure(figsize=(14, 8))
    # p_Rcomp.to_csv('//netapp03.vgt.vito.be/After/Minimal_example_VITO/python/output_plot.csv', index=False)

    # Plotting using matplotlib
    plt.figure(figsize=(12, 8))
    plt.stackplot(
        p["Date"],
        p["stengel en blad"],
        colors=["lightgreen"],
        labels=["stengel en blad"],
    )

    # Adding points for 'stikstof' and 'maai'
    plt.scatter(
        bollen["Date"], bollen["stikstof"], color="orange", s=100, label="Stikstof"
    )
    plt.scatter(bollen["Date"], bollen["maai"], color="darkgreen", s=100, label="Maai")

    plt.xlabel("Datum")
    plt.ylabel("Oogstbare biomassa (kg ha$^{-1}$)")
    plt.legend()
    plt.title("Grass Yield Plot")
    plt.savefig(outdir + "/grass_plot_biomass.png", bbox_inches="tight")


def lingraplot_cut_biomass_composition(lingra_data, outdir):
    os.makedirs(outdir, exist_ok=True)
    # Renaming the columns based on the provided dictionary
    kols = {
        "AG": "Date",
        "AS": "Harvest date",
        "IG": "Fertilizer N application; (0=NO, 1=YES)",
        "KG": "N concentration in total vegetative living above ground biomass",
        "EQ": "Total harvested weight"
    }
    lingra_data.rename(columns=kols, inplace=True)

    # Calculating PY_AGB and Y_wtht_P
    lingra_data['PY_AGB'] = lingra_data['N concentration in total vegetative living above ground biomass'] * \
                            lingra_data['Total harvested weight'] * 6.25
    lingra_data['Y_wtht_P'] = lingra_data['Total harvested weight'] - lingra_data['PY_AGB']

    # Selecting relevant columns and converting the Date column
    lingra_data['Date'] = pd.to_datetime(lingra_data['Date'], unit='D')
    lingra_data = lingra_data[['Date', 'Harvest date', 'Fertilizer N application; (0=NO, 1=YES)', 'PY_AGB', 'Y_wtht_P']]
    lingra_data.rename(columns={'Harvest date': 'maai', 'Fertilizer N application; (0=NO, 1=YES)': 'stikstof'},
                       inplace=True)

    # Filtering where maai == 1 and adding a cumulative sum column
    lingra_data = lingra_data[lingra_data['maai'] == 1]
    lingra_data['snede'] = lingra_data['maai'].cumsum()

    # Melting the dataframe to long format
    df_data = pd.melt(lingra_data, id_vars=['Date', 'maai', 'stikstof', 'snede'], var_name='parameter')

    # Filtering the data for the plot and renaming the parameter levels
    df_data = df_data[df_data['parameter'].isin(['PY_AGB', 'Y_wtht_P'])]
    df_data['parameter'] = pd.Categorical(df_data['parameter'], categories=['Y_wtht_P', 'PY_AGB'], ordered=True)
    df_data['parameter'] = df_data['parameter'].cat.rename_categories(['Other biomass components', 'Crude protein'])
    df_data.to_csv(os.path.join(outdir, 'gras_plot_biomass_composition.csv'), index=False)
    # Plotting
    plt.figure(figsize=(10, 6))
    sns.barplot(x='snede', y='value', hue='parameter', data=df_data, palette=['darkgreen', 'orange'])
    plt.xlabel('Cut')
    plt.ylabel('Biomass yield per cut (kg ha⁻¹)')
    plt.legend(title='Biomass composition')
    plt.title('Biomass Composition by Cut')
    sns.despine()
    plt.savefig(outdir + "/grass_plot_biomass_composition.png", bbox_inches="tight")
