from dataclasses import dataclass
from typing import List, Dict, Any
from datetime import datetime
from dateutil.parser import isoparse


@dataclass
class FieldData:
    irrigations: List[any]
    fertilisations: List[any]
    harvests: List[any]
    geometry: any
    startDate: datetime
    endDate: datetime
    country: str
    community: str
    
    @classmethod
    def from_request(cls, data: Dict[str, Any]):
        # Parse the year from the endDate
        yearOfEndDate = isoparse(data['endDate']).year        
        # Construct startDate dynamically using the parsed year
        startDate = f"{yearOfEndDate}-01-01T12:00:00.000Z"
                
        return cls(
            irrigations=data['irrigations'],
            fertilisations=[treatment for treatment in data['fertilisations'] if treatment['type'] == 'fertilizer'],
            harvests=data['harvests'],
            geometry=data['geometry'],
            startDate=startDate,
            endDate=data['endDate'], 
            country=data['country'],
            community=data['community']
        )  

        # TODO: should be decided whether we will use field startDate even it is in previous year 
        # or we use first day of the endDate year